from .sender import Sender
from .color import Color

import serial
import time

from threading import Lock


class SerialSender(Sender):
    """
    Sends the color data via the usb cable.
    """

    def __init__(self, port_name="/dev/ttyACM0", baudrate=500000):
        super().__init__()
        self._baudrate = baudrate
        self._port = port_name
        self._lock = Lock()
        self._s = serial.Serial(self.port, self.baudrate)
        self._s.flushInput()

    @property
    def baudrate(self):
        return self._baudrate

    @property
    def port(self):
        return self._port

    def init(self, panel):
        super().init(panel)
        self._sendbuffer = bytearray(3 * self.panel.count + 1)

        # Send command to initialize the panel
        # data = [Sender.CMD_INIT_PANEL, self.panel.columns, self.panel.rows]
        # self._s.write(bytearray(data))

    def update(self):
        """Sends the data via the USB cable to the micro controller.
        
        If gamma correction is enabled for the associated display each byte of 
        the color data is mapped to the gamma corrected value.

        The command value ``Sender.CMD_PAINT_PANEL`` may not be one of the
        data values to be transmitted, because for the arduino sketch this marks
        the beginning of a new frame. In gamma corrected color values this is 
        always true, because the command value is not in the gamma8 mapping table.

        If gamma correction is disabled and the color data contains the command 
        value, the data value is increased by one. For example the color value
        ``(234, 5, 235)`` is mapped to the color ``(235, 5, 235)`` on the fly.
        The value in the display data is not changed.

        .. tip::
            Please read the documentation for 
            `PySerial <https://pythonhosted.org/pyserial/>`_, as not every 
            baudrate is supported on every plattform. Also there you will find
            the naming convention for the ports on different plattforms.

        """
        if not self._s:
            raise ValueError("Not initialized")

        with self._lock:

            self._sendbuffer[0] = Sender.CMD_PAINT_PANEL
            for i, value in enumerate(self.panel.data):
                if self.panel.gamma_correction:
                    self._sendbuffer[i + 1] = Color.gammaCorrection(value)
                else:
                    """
                    Make shure, the CMD_PAINT_PANEL is not part of the color data
                    """
                    self._sendbuffer[i + 1] = (
                        value if value != Sender.CMD_PAINT_PANEL else value + 1
                    )

            if self._s:
                self._s.write(self._sendbuffer)
            else:
                raise ValueError("No serial line")



from .framerateemitter import FramerateEmitter
from .gamepademitter import GamepadEmitter
from .keyboardemitter import KeyboardEmitter
from .serialemitter import SerialEmitter
from .eventqueue import EventEmitter, EventDispatcher, Event


__all__ = [
    "EventDispatcher",
    "EventEmitter",
    "FramerateEmitter",
    "GamepadEmitter",
    "KeyboardEmitter",
    "SerialEmitter",
    "Event",
]



from ledwall.util import TimeDelta
from .eventqueue import EventEmitter, Event

import time


class FramerateEmitter(EventEmitter):
    def __init__(self, framerate):
        # type : (int) -> None
        super().__init__()
        self._seconds_per_frame = 1 / framerate
        self._timer = TimeDelta()
        self._timer.begin()
        self._frame = 1

    def emit(self):
        self.queue.put(
            Event(Event.SYSTEM, "update", {"frame": self._frame}, priority=Event.PRIORITY_HIGH)
        )
        self._frame += 1
        time.sleep(self._seconds_per_frame)



from __future__ import print_function, division

from .sender import Sender
from .color import Color

import socket

from threading import Lock
import time


class UDPSender(Sender):
    """An UDPSender instance sends the frame data via UDP (surprise, surprise). UDP is a 
    connectionless communication model. It is not guaranteed that the client will receive
    the packages in the correct order. 

    The UDPSender needs a server address (name or ip) and a port to send the data to. The
    port defaults to 3548.

    In theory the maximum length of a UPD package is 65,535 bytes, because of the 16bit 
    length field. In the real world there will be additional limitations. For a LED Panel
    with 2048 LEDs (which is not the smallest panel) we need 3*2,048 bytes plus 4 command 
    bytes. So size shouldn't be a limitation.

    The good thing about UDP is, that it is a connectionless communication. So if the server, 
    which is the panel, is temporarily not available, nothing will happen. As soon as the server
    is available again, it will continue to do display the frames.

    Because the packages are send through the network, you will not have a stable or guaranteed
    transmission speed. So there may be a noticable delay between sending and receiving.    

    :param str server: Name or IP of the server to send the framedata to.

    :param int port: Portnumber to send the data to. Defaults to 3548

    """

    def __init__(self, server="localhost", port=3548):
        super().__init__()
        self._server = server
        self._port = port
        self._lock = Lock()

    @property
    def server(self):
        return self._server

    @property
    def port(self):
        return self._port

    def init(self, panel):
        super().init(panel)
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sendbuffer = bytearray(3 * self.panel.count + 4)
        self._sendbuffer[0] = Sender.CMD_PAINT_PANEL  # Write Raw
        self._sendbuffer[1] = 2  # Update
        self._sendbuffer[2] = self.panel.count
        self._sendbuffer[3] = 0  # Reserved

    def update(self):
        if not self._socket:
            raise ValueError("Not initialized")

        with self._lock:

            for i, value in enumerate(self.panel.data):
                if self.panel.gamma_correction:
                    self._sendbuffer[i + 4] = Color.gammaCorrection(value)
                else:
                    self._sendbuffer[i + 4] = value

            if self._socket:
                try:
                    self._socket.sendto(self._sendbuffer, (self._server, self._port))
                    # time.sleep(0.3)
                except Exception as e:
                    print("Could not send data via udp. ", e)
            else:
                raise ValueError("No Socket Connection")



"""
Point Geometry

Copyright 2019

klaas.nebuhr@digicubes.de
"""
from ledwall.components import Display
from ledwall.components.color import TColor


class Point(object):
    """
	Point class.

	A pont has a x and a y component.
	"""

    def __init__(self, x, y):
        self._x = x
        self._y = y

    @property
    def x(self):
        """The x component."""
        return self._x

    @x.setter
    def x(self, value):
        self._x = value

    @property
    def y(self):
        """The y component"""
        return self._y

    @y.setter
    def y(self, value):
        self._y = value

    @property
    def position(self):
        """
		Returns the current position as a tuple
		(x, y).
		"""
        return (self.x, self.y)

    def __len__(self):
        return 2

    def __getitem__(self, key):
        return {0: self.x, 1: self.y,}.get(key, None)

    def __iter__(self):
        yield self.x
        yield self.y

    def __add__(self, value):
        result = Point(self.x, self.y)
        result += value
        return result

    def __radd__(self, value):
        result = Point(self.x, self.y)
        result += value
        return result

    def __iadd__(self, value):
        if isinstance(value, int):
            self._x += value
            self._y += value
            return self

        if isinstance(value, (tuple, list, Point)) and len(value) == 2:
            self._x += value[0]
            self._y += value[1]
            return self

        return NotImplemented

    def __eq__(self, value):
        if isinstance(value, int):
            return self.x == value and self.y == value

        if isinstance(value, (tuple, list, Point)) and len(value) == 2:
            return self.x == value[0] and self.y == value[1]

        return False

    def __ne__(self, value):
        return not (self == value)

    def __str__(self):
        return "({:d},{:d})".format(self.x, self.y)

    def __repr__(self):
        return "Point({:d},{:d})".format(self.x, self.y)

    def __isub__(self, value):

        if isinstance(value, int):
            self._x -= value
            self._y -= value
            return self
        if isinstance(value, (tuple, list, Point)) and len(value == 2):
            print(value[0])
            self._x -= value[0]
            self._y -= value[1]
            return self

        raise ValueError("Cannot add value", value)

    def paint(self, display: Display, color: TColor = None, update: bool = False):
        """
		Writes the point to the given display.
		"""
        if color is None:
            color = (255, 255, 255)

        # TODO: Check boundaries
        display.set_pixel(self.x, self.y, color, update)



from .sender import Sender


class ConsoleSender(Sender):
    def __init__(self):
        super().__init__()

    def init(self, panel):
        super().init(panel)

    def update(self):
        print("Panel {} in frame nr.{:d}".format(self.panel.id, self.panel.frame))



from .color import Color


class ColorTable(object):

    AliceBlue = Color.fromHexString("#F0F8FF")
    Amethyst = Color.fromHexString("#9966CC")
    AntiqueWhite = Color.fromHexString("#FAEBD7")
    Aqua = Color.fromHexString("#00FFFF")
    Aquamarine = Color.fromHexString("#7FFFD4")
    Azure = Color.fromHexString("#F0FFFF")
    Beige = Color.fromHexString("#F5F5DC")
    Bisque = Color.fromHexString("#FFE4C4")
    Black = Color.fromHexString("#000000")
    BlanchedAlmond = Color.fromHexString("#FFEBCD")
    Blue = Color.fromHexString("#0000FF")
    BlueViolet = Color.fromHexString("#8A2BE2")
    Brown = Color.fromHexString("#A52A2A")
    BurlyWood = Color.fromHexString("#DEB887")
    CadetBlue = Color.fromHexString("#5F9EA0")
    Chartreuse = Color.fromHexString("#7FFF00")
    Chocolate = Color.fromHexString("#D2691E")
    Coral = Color.fromHexString("#FF7F50")
    CornflowerBlue = Color.fromHexString("#6495ED")
    Cornsilk = Color.fromHexString("#7FFF00")
    Chartreuse = Color.fromHexString("#FFF8DC")


"""
  Crimson =0xDC143C, Cyan =0x00FFFF, DarkBlue =0x00008B, DarkCyan =0x008B8B,
  DarkGoldenrod =0xB8860B, DarkGray =0xA9A9A9, DarkGrey =0xA9A9A9, DarkGreen =0x006400,
  DarkKhaki =0xBDB76B, DarkMagenta =0x8B008B, DarkOliveGreen =0x556B2F, DarkOrange =0xFF8C00,
  DarkOrchid =0x9932CC, DarkRed =0x8B0000, DarkSalmon =0xE9967A, DarkSeaGreen =0x8FBC8F,
  DarkSlateBlue =0x483D8B, DarkSlateGray =0x2F4F4F, DarkSlateGrey =0x2F4F4F, DarkTurquoise =0x00CED1,
  DarkViolet =0x9400D3, DeepPink =0xFF1493, DeepSkyBlue =0x00BFFF, DimGray =0x696969,
  DimGrey =0x696969, DodgerBlue =0x1E90FF, FireBrick =0xB22222, FloralWhite =0xFFFAF0,
  ForestGreen =0x228B22, Fuchsia =0xFF00FF, Gainsboro =0xDCDCDC, GhostWhite =0xF8F8FF,
  Gold =0xFFD700, Goldenrod =0xDAA520, Gray =0x808080, Grey =0x808080,
  Green =0x008000, GreenYellow =0xADFF2F, Honeydew =0xF0FFF0, HotPink =0xFF69B4,
  IndianRed =0xCD5C5C, Indigo =0x4B0082, Ivory =0xFFFFF0, Khaki =0xF0E68C,
  Lavender =0xE6E6FA, LavenderBlush =0xFFF0F5, LawnGreen =0x7CFC00, LemonChiffon =0xFFFACD,
  LightBlue =0xADD8E6, LightCoral =0xF08080, LightCyan =0xE0FFFF, LightGoldenrodYellow =0xFAFAD2,
  LightGreen =0x90EE90, LightGrey =0xD3D3D3, LightPink =0xFFB6C1, LightSalmon =0xFFA07A,
  LightSeaGreen =0x20B2AA, LightSkyBlue =0x87CEFA, LightSlateGray =0x778899, LightSlateGrey =0x778899,
  LightSteelBlue =0xB0C4DE, LightYellow =0xFFFFE0, Lime =0x00FF00, LimeGreen =0x32CD32,
  Linen =0xFAF0E6, Magenta =0xFF00FF, Maroon =0x800000, MediumAquamarine =0x66CDAA,
  MediumBlue =0x0000CD, MediumOrchid =0xBA55D3, MediumPurple =0x9370DB, MediumSeaGreen =0x3CB371,
  MediumSlateBlue =0x7B68EE, MediumSpringGreen =0x00FA9A, MediumTurquoise =0x48D1CC, MediumVioletRed =0xC71585,
  MidnightBlue =0x191970, MintCream =0xF5FFFA, MistyRose =0xFFE4E1, Moccasin =0xFFE4B5,
  NavajoWhite =0xFFDEAD, Navy =0x000080, OldLace =0xFDF5E6, Olive =0x808000,
  OliveDrab =0x6B8E23, Orange =0xFFA500, OrangeRed =0xFF4500, Orchid =0xDA70D6,
  PaleGoldenrod =0xEEE8AA, PaleGreen =0x98FB98, PaleTurquoise =0xAFEEEE, PaleVioletRed =0xDB7093,
  PapayaWhip =0xFFEFD5, PeachPuff =0xFFDAB9, Peru =0xCD853F, Pink =0xFFC0CB,
  Plaid =0xCC5533, Plum =0xDDA0DD, PowderBlue =0xB0E0E6, Purple =0x800080,
  Red =0xFF0000, RosyBrown =0xBC8F8F, RoyalBlue =0x4169E1, SaddleBrown =0x8B4513,
  Salmon =0xFA8072, SandyBrown =0xF4A460, SeaGreen =0x2E8B57, Seashell =0xFFF5EE,
  Sienna =0xA0522D, Silver =0xC0C0C0, SkyBlue =0x87CEEB, SlateBlue =0x6A5ACD,
  SlateGray =0x708090, SlateGrey =0x708090, Snow =0xFFFAFA, SpringGreen =0x00FF7F,
  SteelBlue =0x4682B4, Tan =0xD2B48C, Teal =0x008080, Thistle =0xD8BFD8,
  Tomato =0xFF6347, Turquoise =0x40E0D0, Violet =0xEE82EE, Wheat =0xF5DEB3,
  White =0xFFFFFF, WhiteSmoke =0xF5F5F5, Yellow =0xFFFF00, YellowGreen =0x9ACD32,
  FairyLight =0xFFE42D, FairyLightNCC =0xFF9D2A
  """



from __future__ import print_function

from .sender import Sender
from .color import Color

import paho.mqtt.client as mqtt
from socket import error as SocketError


def on_connect(client, userdata, flags, rc):
    print("Connected with result code " + str(rc))
    client.subscribe("$SYS/#")
    client.loop_start()


def on_message(client, userdata, msg):
    print(msg.topic)


class MqttSender(Sender):
    """

    """

    MAX_PAYLOAD_SIZE = 80
    """The maximum length for the payload part of mqtt 
    message in bytes. This includes the command part as well
    as the data patr of every command.
    """

    def __init__(self, server="nebuhr", port=1883):
        """Server is the server name or the ip of the mqtt server.
        This class was successfully tested with the mosquitto.

        :param str server: Name or IP of the MqttServer.
        :param int port: The port of the MqttServer
        """
        super().__init__()
        self._server = server
        self._port = port
        self._client = mqtt.Client()

        self._client.on_connect = on_connect
        self._client.on_message = on_message

        self._client.connect(server, port, 60)

    # The callback for when a PUBLISH message is received from the server.

    def update(self):
        max_payload_size = MqttSender.MAX_PAYLOAD_SIZE - 7
        data = self.panel.data
        panel_data_size = len(data)

        for i in range(0, panel_data_size, max_payload_size):
            self.raw_write(i, [Color.gammaCorrection(x) for x in data[i : i + max_payload_size]])

        self.raw_show()

    def _set_pixel(self, offset, r, g, b):
        self._publish([Sender.CMD_SET_PIXEL] + self.itob(offset) + self._frame_number + [r, g, b])

    def raw_write(self, offset, data):
        size = len(data)
        if offset >= self.panel.byte_count:
            raise ValueError("Offset to high")
        if offset < 0:
            raise ValueError("Offset may not be negative")
        if (offset + size) > self.panel.byte_count:
            raise ValueError("Data exceeds buffer size")

        self._publish(
            [Sender.CMD_WRITE_RAW]
            + self._frame_number
            + self.itob(offset)
            + self.itob(size)
            + [x for x in data]
        )

    def raw_show(self):
        self._publish([Sender.CMD_SHOW] + self._frame_number)

    def init(self, panel):
        """

        :param panel:
        :return:
        """
        super().init(panel)
        data = [Sender.CMD_INIT_PANEL, self.panel.columns, self.panel.rows] + self._frame_number
        self._publish(data)

    def _publish(self, data):
        try:
            self._client.publish(self.panel.id, bytearray(data))
        except SocketError:
            print("Could not send data.")

    @staticmethod
    def itob(value):
        return [(value >> 8) & 0xFF, value & 0xFF]

    @property
    def _frame_number(self):
        return self.itob(self.panel.frame)



import pyowm

"""
  See: https://github.com/csparpa/pyowm
"""


class Weather(object):
    def __init__(self, settings):
        self._cfg = settings["weather"]
        self._owm = pyowm.OWM(self._cfg["API_key"])

    @property
    def owm(self):
        return self._owm

    @property
    def observation(self):
        return self.owm.weather_at_place("Bremen,DE")

    @property
    def weather(self):
        return self.observation.get_weather()



import json


class Settings(object):
    def __init__(self, filename="settings.json"):
        self._filename = filename
        with open(filename, "r") as f:
            self._jsondata = json.load(f)

    def __str__(self):
        return json.dumps(self._jsondata, sort_keys=True, indent=4)

    def __repr__(self):
        return "Settings(filename='%s')" % self._filename

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._jsondata[key]
        return NotImplemented



"""
Dimension class.

A dimension is a data structure containing
the two components width and height.
"""


class Dimension(object):
    def __init__(self, width: int, height: int):
        self._width = int(width)
        self._height = int(height)

    @property
    def height(self) -> int:
        """
		The height component.
		"""
        return self._height

    @property
    def width(self) -> int:
        """
		The height component.
		"""
        return self._width

    def __iter__(self):
        yield self._width
        yield self._height

    def __len__(self):
        return 2

    def __repr__(self):
        return "Dimension({:d},{:d})".format(self._width, self._height)

    def __str__(self):
        return "({:d},{:d})".format(self._width, self._height)



from __future__ import division, print_function

import sys
import itertools
import time
from typing import Optional, List

from enum import IntEnum

from .color import Color, TColor
from .sender import Sender
from ..util import TimeDelta
from ..geometry import Rectangle, Point
from .asyncsender import AsyncSender

PIL_AVAILABLE = True

try:
    from PIL import Image
except ImportError:
    Image = None
    print("Python Image library (PIL) not availabe. Image functions will be disabled")

BYTES_PER_PIXEL = 3


class WireMode(IntEnum):
    """
    The wire mode describes the way, the leds are organized on the board. Two ways of wiring are supported.
    *LTR* and *ZIGZAG*. The WireMode class is an Enum. Whenever you have to specifiy a mode, you simply write
    ``WireMode.LTR`` or ``WireMode.ZIGZAG`` 

    **WireMde.LTR  (Left-To-Right)**

    In the mode WireMode.LTR, all rows are going from left to right. The cable for the data line (blue) goes 
    all the way back from the end of one row to the beginning of the next row.

    .. figure:: mode_ltr.png
       :scale: 60 %
       :alt: mode ltr
       :align: center

       WireMode.LTR

    **WireMode.ZIGZAG**

    In the mode WireMode.ZIGZAG all even rows go from left to right where as the the odd rows go from right to left. Organizing leds
    in this way saves a lot of cable for the data line (blue).

    .. figure:: mode_zigzag.png
       :scale: 60 %
       :alt: mode ltr
       :align: center

       WireMode.ZIGZAG

    """

    LTR = 0
    ZIGZAG = 1


class Display(object):
    """Creates a new instance of a led display. The Display class manages a virtual representation of the color state 
    of the LEDs on the physical panel. The class offers many methods to set and change colors. The physical LEDs are updated via the
    :meth:`~ledwall.components.Display.update` method. Because there are different ways to connect the arduino (or ESP) to your
    computer, the transmission of the data is managed by an instance of a :class:`~ledwall.components.Sender`.
    This library offers several Implementations (:class:`~ledwall.components.SerialSender`,
    :class:`~ledwall.components.MqttSender`, :class:`~ledwall.components.UDPSender`).

    The methods to manipulate the color state of the pixels pay respect to the wiring mode you used. There are two ways to
    layout of LEDs on the board specified by the WireMode parameter in the constructor. 


    A very basic program could look like this:

    .. code-block:: python
    
        from ledwall.components import *

        s = SerialSender()            # Creating a serial sender with the default port_name and baudrate
        d = Display(16,32,s)          # Defining a new display component
        
        red   = Color(255,0,0)        # Defining an RGB color
        col   = HSVColor(0.7,0.8,1.0) # Defining an HSVColor    

        d.fill(green)
        d.set_pixel(0,3,red)          # which is equivalent to d[(0,3)] = red or d[(0,3)] = (255,0,0)
        d.set_pixel(14,23,col)

        col.hue += 0.13               # Changing the HUE component of the color
        d.set_pixel(15,23,col)        # Setting the pixel in this color

        d.update()                    # Updating the physical component

    :param cols: The number of columns of the display
    :type cols: int

    :param rows: The number of rows of the display.
    :type rows: int

    :param sender: Instance of the sender (One of the subclasses, to be more precise)
    :type sender: Sender 

    :param framerate: The intended number of frames per second.
    :type framerate: int

    :param mode: The mode that the LEDs are organized. Left-to-Right or Zig-Zag. Defaults to WireMode.LTR.
    :type mode: WireMode

    :param panel_id: An unique id for this panel. Currently not used. 
    :type panel_id: str

    :param async_mode: If True, the update() method of the provided sender will be called asynchronously.
    :type async_mode: boolean

    :rtype: None
    """

    def __init__(
        self,
        cols: int,
        rows: int,
        sender: Optional[Sender] = None,
        framerate: Optional[int] = None,
        mode: WireMode = WireMode.LTR,
        panel_id: str = "LEDPANEL",
        async_mode: bool = False,
    ):
        self._cols = int(cols)
        self._rows = int(rows)
        self._data = [0] * (BYTES_PER_PIXEL * self.count)
        self._mode = mode
        self._transmissionTime = TimeDelta()
        self._frame_nr = 0
        self._gamma_correction = True
        self._id = panel_id
        self._sender = sender
        self._framerate = framerate
        self._millis_per_frame = 1000 / framerate if framerate else 0
        self._frame_computation_time = TimeDelta()

        if sender and async_mode:
            self._sender = AsyncSender(sender)

        if self._cols < 1:
            raise ValueError("Argument cols must be a value greater than 1.", cols)

        if self._rows < 1:
            raise ValueError("Argument rows must be a value greater than 1.", cols)

        if self._sender:
            self._sender.init(self)

    @property
    def mode(self):
        """
        Returns the mode for this display. The return type is
        :class:`~ledwall.components.WireMode`
        """
        return self._mode

    @property
    def data(self):
        """
        Returns the raw byte array for the color data. 
        """
        return self._data

    @property
    def fct(self):
        """
        The frame computation time.
        """
        return self._frame_computation_time

    @property
    def id(self):
        """
        The panel id as set in the constructor. (read-only)
        """
        return self._id

    def __iter__(self):
        index = 0
        while index < self.count:
            yield tuple(self._data[index * BYTES_PER_PIXEL : (index + 1) * BYTES_PER_PIXEL])
            index += 1

    def __len__(self):
        return self._cols * self._rows

    def __getitem__(self, key):
        if isinstance(key, (tuple, list)) and len(key) == 2:
            index = self._coords_to_index(key[0], key[1]) * 3
            return tuple(self._data[index : index + 3])

        if isinstance(key, int):
            index = key * 3
            return tuple(self._data[index : index + 3])

        if isinstance(key, slice):
            return [color for color in itertools.islice(self, key.start, key.stop, key.step)]

        return NotImplemented

    def _set_color_at(self, index: int, color: TColor):
        if index >= self.count:
            raise ValueError(
                "Index out of range. Maximum is %d but was %d" % (self.count - 1, index)
            )

        index *= BYTES_PER_PIXEL
        color = Color.convert(color)
        self._data[index] = color.red
        self._data[index + 1] = color.green
        self._data[index + 2] = color.blue
        return

    def set_colors(self, colors: List[TColor], transparent_color: Optional[TColor] = None):
        """
        Sets the colors of the pixels according to the
        colors in the color list parameter.

        The method ignores the wiring of the display and
        sets the colors in the given order. The result will
        look different on differently wired display.

        Any entry in the colors list, where `colors[i] == transparent_color`
        is `True` will be ignored and the corresponding display pixel will
        be skipped.
        """

        index = 0
        for c in colors:
            if c and c != transparent_color:
                self._set_color_at(index, c)
            index += 1

    def get_colors(self) -> List[TColor]:
        return self[:]

    def __setitem__(self, key, item: TColor):
        if not item:
            raise ValueError("None is not allowed for item. Item must be a color instance")

        if isinstance(key, int):
            if key < 0:
                raise ValueError("Index may not below zero.", key, item)

            if key >= self.count:
                raise ValueError("Index must below count.", key, item)

            self._set_color_at(self._adjust_index(key), item)
            return

        if isinstance(key, (tuple, list)) and len(key) == 2:
            self._set_color_at(self._coords_to_index(key[0], key[1]), item)

        if isinstance(key, slice):
            i = key.start or 0
            stop = key.stop or self.count
            step = key.step or 1
            while i < stop:
                if self._set_color_at(i, item) == NotImplemented:
                    return
                i += step
            return

        return NotImplemented

    @property
    def columns(self) -> int:
        """
        :return: The number of columns on this display
        :rtype: int
        """
        return self._cols

    @property
    def rows(self) -> int:
        """
        :return: The number of rows on this display
        :rtype: int
        """
        return self._rows

    @property
    def count(self) -> int:
        """
        :return: The number of pixels on this display
        :rtype: int
        """
        return self.columns * self.rows

    @property
    def frame(self) -> int:
        """
        :return: The current frame number
        :rtype: int
        """
        return self._frame_nr

    @property
    def byte_count(self) -> int:
        """
        The number of bytes needed for one frame.
        This is equivalent to (width * height * 3).
        """
        return len(self._data)

    @property
    def transmission_info(self):
        return self._transmissionTime.asTuple()

    @property
    def gamma_correction(self) -> bool:
        """
        Is gamma correction enabled or not.
        """
        return self._gamma_correction

    @gamma_correction.setter
    def gamma_correction(self, value):
        self._gamma_correction = value

    @property
    def frame_rate(self):
        """Returns the framerate."""
        return self._framerate

    def _test_coords(self, x, y):
        """
        Returns true, if x and y are valid indices for
        a pixeln on tis display.

        :param int x: X position
        :param int y: Y position
        :return: `True` if `0 <= x < self.columns and 0 <= y < self.rows`.
            False else.
        :rtype: bool
        """
        return 0 <= x < self.columns and 0 <= y < self.rows

    def _adjust_column(self, x, y):
        """

        :param int x: X position
        :param int y: Y position
        :return: The adjusted column index.
        """
        if self._mode == WireMode.ZIGZAG and self.odd_row(y):
            return self.columns - x - 1

        return x

    def _adjust_index(self, index):
        x, y = self._index_to_coords(index, False)
        return self._coords_to_index(x, y)

    def _coords_to_index(self, x, y, adjust=True):
        if adjust:
            return (y * self.columns) + self._adjust_column(x, y)
        return (y * self.columns) + x

    def _index_to_coords(self, index, adjust=True):
        x = index % self.columns
        y = index // self.columns
        if adjust:
            return self._adjust_column(x, y), y
        return x, y

    def set_pixel(self, x: int, y: int, color: TColor, update: Optional[bool] = False):
        """Setting the color for a single pixel. 

        Returns True, if setting the pixel was successful, Fales else. The color is specified by a color instance.
        The method also accepts tuples or arrays as colors. The length has to be at least 3 and must contain the
        value for red, green, and blue.

        The following lines have the same effect.

        .. code-block:: python

            d.set_pixel(0,3,red,True)
            d.set_pixel(0,3,(255,0,0),True)
            d.set_pixel(0,3,[255,0,0],True)
            
        :param x: The x position of the pixel. Must be a value with the following constraint: ``0 <= x < width``
        :type x: int

        :param y: The number of rows of the display.
        :type y: int

        :param color: 
            The color for the pixel. If you want to deactivate or clear a pixel, just use black (0,0,0)
            as color value.
        :type color: Color, tuple, list

        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: boolean
        """
        if not self._test_coords(x, y):
            return False

        self._set_color_at(self._coords_to_index(x, y), color)
        self.update(update)
        return True

    def get_pixel(self, x, y):
        """Getting the color for a single pixel. 

        Returns the color for the pixel located at (x,y). The x value must be within the the range: ``0 <= x < width``. 
        The y value must be with the range: ``0 <= y < height``

        :param x: The x position of the pixel. Must be a value with the following constraint: ``0 <= x < width``
        :type x: int

        :param y: The number of rows of the display.
        :type y: int

        :rtype: Color
        """
        assert self._test_coords(
            x, y
        ), "Coords out of range ({},{}). Dimensions of the 0-indexed display are ({},{})".format(
            x, y, self.columns, self.rows
        )

        index = self._coords_to_index(x, y) * 3
        return Color.fromTuple(tuple(self._data[index : index + 3]))

    def write_bitmask(self, row, value, color1=(266, 165.0), color0=(0, 0, 0)):
        index = self._coords_to_index(6, row)
        for _ in range(self.columns):
            self._set_color_at(index, color1 if value & 1 else color0)
            value >>= 1
            index -= 1

    def horizontal_line(self, row, color, update=False):
        """Drawing a horizontal line in the specified color. 

        The row value must be with the range: ``0 <= row < height``

        :param row: The row to be filled.
        :type row: int

        :param color: The color for the row. If you want to deactivate or clear a row, just use black (0,0,0)
            as color value.
        :type color: Color, tuple, list

        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: None
        """
        if 0 > row >= self.rows:
            raise ValueError("Row index out of bounds.", row)

        self[row * self.columns : ((row + 1) * self.columns)] = Color.convert(color)
        self.update(update)

    def vertical_line(self, column, color, update=False):
        """Drawing a vertical line in the specified color. See also :meth:`~ledwall.components.Display.horizontal_line`

        The column value must be with the range: ``0 <= column < width``

        :param column: The column to be filled.
        :type column: int

        :param color: The color for the column. If you want to deactivate or clear a row, just use black (0,0,0)
            as color value.
        :type color: Color, RGBColor, HSVColor, tuple, list
        
        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: None
        """
        assert 0 <= column < self.columns

        color = Color.convert(color)

        for i in range(self.rows):
            self._set_color_at(self._coords_to_index(column, i), color)

        self.update(update)

    @staticmethod
    def odd_row(row):
        """

        :param int row: Test, if the index of the row is even (False) or odd (True)
        :return: True, if row is odd, False else.
        """
        return (row & 1) == 1

    def shift_row_left(self, row, update=False):
        """Shifts row one pixel to the left. The outermost left pixel will be plced to the right.
        So actually it rotates the pixels in the row. The method pays respect to the 
        :attr:`~ledwall.components.Display.mode` property.

        :param int row: Test, if the index of the row is even (False) or odd (True)
        """
        start_index = row * self.columns * 3

        if self.odd_row(row) and self._mode == WireMode.ZIGZAG:
            width = self.columns * 3
            right_pixel = self._data[start_index + width - 3 : start_index + width]
            tmp = self._data[start_index : start_index + width - 3]
            self._data[start_index + 3 : start_index + width] = tmp
            self._data[start_index : start_index + 3] = right_pixel[:]

        else:
            width = self.columns * 3
            left_pixel = self._data[start_index : start_index + 3]
            self._data[start_index : start_index + width - 3] = self._data[
                start_index + 3 : start_index + width
            ]
            self._data[start_index + width - 3 : start_index + width] = left_pixel[:]

        self.update(update)

    def shift_left(self, update=False):
        """Shifts all pixels to the left.

        The method shift pixels to the left visually, because the method takes the mode into account. This means, if
        the mode is :attr:`~ledwall.components.Display.MODE_ZIGZAG` and the row number is odd, then the pixels are shifted physically to the
        right. But because this row reads left to right, the pixels are shifted visually to the left.

        The pixel to the outermost left will be placed at the last column. So this method implements a rotation of
        the pixels.

        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: None
        """
        for row in range(self.rows):
            self.shift_row_left(row)

        self.update(update)

    def move(self, count=1):
        """Moves all pixels in the buffer. 

        .. warning::
            This method does not take the ``mode`` into account. So the visual effect differs depending on the mode.

        :param int count: The amount each pixel is moved.
        """
        self._data[BYTES_PER_PIXEL * count :] = self._data[: -BYTES_PER_PIXEL * count]

    def shift_row_right(self, row, update=False):
        """Shifts all pixels in the specified row to the right.

        The method shift pixels to the right visually, because the method takes the mode into account. This means, if
        the mode is :attr:`~ledwall.components.Display.MODE_ZIGZAG` and the row number is odd, then the pixels are 
        shifted physically to the left. But because this row reads left to right, the pixels are shifted visually 
        to the right.

        The pixel to the outermost right will be placed at the first column. So this method implements a rotation of
        the pixels.

        :param row: The number of the row to be shifted. The row value must be with the range: ``0 <= row < height``
        :type row: int

        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: None
        """
        start_index = row * self.columns * 3

        if self.odd_row(row) and self._mode == WireMode.ZIGZAG:
            width = self.columns * 3
            left_pixel = self._data[start_index : start_index + 3]
            self._data[start_index : start_index + width - 3] = self._data[
                start_index + 3 : start_index + width
            ]
            self._data[start_index + width - 3 : start_index + width] = left_pixel[:]

        else:
            width = self.columns * 3
            right_pixel = self._data[start_index + width - 3 : start_index + width]
            tmp = self._data[start_index : start_index + width - 3]
            self._data[start_index + 3 : start_index + width] = tmp
            self._data[start_index : start_index + 3] = right_pixel[:]

        self.update(update)

    def shift_right(self, update=False):
        """Shifts all pixels to the right.

        The method shift pixels to the right visually, because the method takes the mode into account. This means, if
        the mode is :attr:`~ledwall.components.Display.MODE_ZIGZAG` and the row number is odd, then the pixela are 
        shifted physically to the left. But because this row reads left to right, the pixels are shifted visually to 
        the right.

        The pixel to the outermost right will be placed at the first column. So this method implements a rotation of
        the pixels.

        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: None
        """
        for row in range(self.rows):
            self.shift_row_right(row)

        self.update(update)

    def as_rectangle(self):
        return Rectangle(0, 0, self.columns, self.rows)

    @property
    def bounds(self):
        """Returns a Rectangle instance where width is equal
        to the number of columns and height ist equal to the 
        number of rows.

        :return: Rectangle with a size equal to the size of this display.
        :rtype: Rectangle
        """
        return Rectangle(0, 0, self.columns, self.rows)

    def fill_rect(self, x, y, w, h, color, update=False):
        """Fills a rectangle in the specified color
        
        :param int x: X position of the top left corner
        
        :param int y: Y position of the top left corner
        
        :param int w: Width of the rectangle
        
        :param int h: Height of the rectangle
        
        :param color: The color for the rectangle. If you want to deactivate or clear a rectangle, just use black (0,0,0) as color value.
        :type color: Color, tuple, list
        
        :param update: If True, the display will be updated.
        :type update: boolean
        
        :rtype: None
        """
        rect = Rectangle(x, y, w, h)
        rect = self.as_rectangle() - rect

        if rect:
            rect = Rectangle.fromTuple(rect)
            for px in range(rect.x, rect.right + 1):
                for py in range(rect.y, rect.bottom + 1):
                    self.set_pixel(px, py, color)
        else:
            print("No intersection found")

        self.update(update)

    def fill(self, color, update=False):
        """Fills the LED display in the specified color.

        :param color: The color for the display. If you want to deactivate or clear the panel, just use
            black (0,0,0) as color value.
        :type color: Color, tuple, list
        
        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: None
        """
        color = Color.convert(color)
        self._data[::3] = [color.red] * self.count
        self._data[1::3] = [color.green] * self.count
        self._data[2::3] = [color.blue] * self.count
        self.update(update)

    def clear(self, update=False):
        """Clears the display (sets all pixel to black)

        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: None
        """
        self._data[:] = [0] * (BYTES_PER_PIXEL * self.count)
        self.update(update)

    def update(self, update=True):
        """Updates the LED display

        The internal frame buffer is send to the arduino board and the LED stripe gets updated.

        The method takes framerate into account. If the time between this update and the last update
        is less then the milliseconds per frame, the method will sleep 'til the end of the frame and 
        update the data at the end of the frame.

        Every time the LED display is updated, the frame number will be increased by one.

        .. note::
           If you are interested to see the time consumed by transferring the date to the arduino,
           you can read the transmissionInfo property after updating. The following code snippet shows
           the basic usage:

        Read out the timings for transmission:

        .. code-block:: python

            d.fill(Color(255,0,0))
            print d.transmission_info

        :param update: If True, the display will be updated.
        :type update: boolean

        :rtype: None
        """
        if not update:
            return

        self._frame_nr += 1

        if self._sender:
            self._sender.update()

        if self._framerate:
            time.sleep(1 / self._framerate)

    def show_image(self, path, update=False, transparent_color=None):
        self.load_image(path, update, transparent_color)

    def load_image(self, path, update=False, transparent_color=None):
        """Loads an image into the LED buffer.

        The method loads an image located at *path* into the LED buffer. If a transparent
        color is provided, all *transparent* colors are ignored. If *update = True* the panel
        data will be updated immediately.

        :param path: The relative or absolute path to the image. The method uses the PIL library
            for reading the file. So any file type supported by PIL is supported by this method.
        :type path: str

        :param update: If True, the panel gets updated via the sender component. Defaults to False
        :type update: boolean
        
        :param transparent_color: The color which defines transparency. Defaults to None.
        :type transparent_color: tuple
        """
        if "PIL" not in sys.modules:
            raise ValueError(
                "Module PIL not available. Consider to install PIL to use this function."
            )
        img = Image.open(path)
        rgbimg = img.convert("RGB")
        for y in range(self.rows):
            for x in range(self.columns):
                color = rgbimg.getpixel((x, y))
                if transparent_color != color:
                    self.set_pixel(x, y, color)
        self.update(update)

    def copy_region_from(
        self, src, rect_src=None, point_dst=Point(0, 0), transparent_color=None, update=False
    ):
        """Copy a region from another display. If a transparent color is provided, all pixels in the
        source panel with the corresponding color will be ignored. 

        The region specified by the :class:`~ledwall.geometry.Rectangle` *rectSrc* will be copied to 
        position specified by *pointDst*. The parameter *pointDst* defaults to the upper left corner 
        :class:`~ledwall.geometry.Point` (0,0).

        :param Display src: The source display to copy the color values from.

        :param Rectangle rect_src: The position and the size of the region to copy. 
            If no value is provided, it defaults to the size of self and position 0,0

        :param Point point_dst: Where to copy to. Defaults to (0,0)

        :param Color transparent_color: Color to ignore while copying. Defaults to None

        :param boolean update: Update display after this operation. Defaults to False
        """
        if not rect_src:
            rect_src = Rectangle(0, 0, self.columns, self.rows)

        for x in range(rect_src.width):
            for y in range(rect_src.height):
                color = src.get_pixel(rect_src.x + x, rect_src.y + y)
                if color and color != transparent_color:
                    self.set_pixel(point_dst.x + x, point_dst.y + y, color)
        self.update(update)



class ColorMap:
    def __init__(self):
        self.__colormap = {}

    def __len__(self):
        return len(self.__colormap)

    def add_entry(self, key, value):
        self.__colormap[key] = value

    def __iadd__(self, other):
        self.__colormap[other[0]] = other[1]
        return self

    def __isub__(self, other):
        if other[0] in self.__colormap:
            del self.__colormap[other[1]]
        return self

    def __getitem__(self, key):
        return self.__colormap.get(key)


class Sprite:
    def __init__(self, pixel, colormap):
        self.__pixel = pixel
        self.__colormap = colormap

    @property
    def colormap(self):
        return self.__colormap

    @property
    def width(self):
        return len(self.__pixel[0])

    @property
    def height(self):
        return len(self.__pixel)

    def paint(self, display, dx, dy):
        if not self.colormap:
            return

        for y, row in enumerate(self.__pixel):
            for x, key in enumerate(row):
                color = self.colormap[key]
                if color:
                    display.set_pixel(x + dx, y + dy, color)



from datetime import datetime


class Clock(object):
    def __init__(self):
        self._x = 0
        self._y = 0

        self._hourColorHigh = (128, 0, 128)
        self._hourColorLow = (5, 0, 5)

        self._minuteColorHigh = (0, 128, 128)
        self._minuteColorLow = (0, 5, 5)

        self._secondColorHigh = (0, 128, 0)
        self._secondColorLow = (0, 5, 0)

        self._display = None

    @property
    def hourColor(self):
        return (self._hourColorHigh, self._hourColorLow)

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        self._x = value

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        self._y = value

    @property
    def position(self):
        return (self.x, self.y)

    @property
    def display(self):
        return self._display

    @display.setter
    def display(self, display):
        self._display = display

    def update(self, update):
        pass


class BinaryCodedSexagesimalClock(Clock):
    def __init__(self, x, y):
        Clock.__init__(self)
        self.x = x
        self.y = y

    def update(self, update=False):
        display = self.display
        if not display:
            return

        now = datetime.now()
        display.write_bitmask(self.y, now.hour, self._hourColorHigh, self._hourColorLow)
        display.write_bitmask(self.y + 1, now.minute, self._minuteColorHigh, self._minuteColorLow)
        display.write_bitmask(self.y + 2, now.second, self._secondColorHigh, self._secondColorLow)
        display.update(update)



from .hsvcolor import HSVColor
from .rgbcolor import RGBColor


def rgb_to_hsv(rgb):
    """Converts a color from RGB to HSV colorspace.

	:param rgb: The color to be converted. Accepted types are list, tuple and RGBColor. If a list or a tuple is provided, 
		the values must be int values or at least covertable to int. The values have to be within the range [0,255]

	:type rgb: (Color, list, tuple)
	"""
    if isinstance(rgb, RGBColor):
        return HSVColor(*rgb.hsv)
    if isinstance(rgb(list, tuple)) and len(rgb) == 3:
        return HSVColor(RGBColor(*rgb).hsv)


def hsv_to_rgb(hsv):
    """Converts a color from HSV to RGB colorspace.

	:param rgb: The color to be converted. Accepted types are list, tuple and Color. If a list or a tuple is provided, 
		the values must be int values or at least covertable to int. The hue value has to be within the range [0,360[.
		The values for saturation and value have to be within the range [0,100]

	:type rgb: (Color, list, tuple)
	"""
    if isinstance(hsv, HSVColor):
        return RGBColor(*hsv.rgb)
    if isinstance(hsv(list, tuple)) and len(hsv) == 3:
        return RGBColor(*(HSVColor(*hsv).rgb))



from .sender import Sender
from .color import Color

import socket

from threading import Lock
import time


class TCPSender(Sender):
    def __init__(self, server="localhost", port=3548):
        super().__init__()
        self._server = server
        self._port = port
        self._lock = Lock()

    @property
    def server(self):
        return self._server

    @property
    def port(self):
        return self._port

    def init(self, panel):
        super().init(panel)
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self.server, self.port))
        self._sendbuffer = bytearray(3 * self.panel.count + 1)
        print("Creating bytearray of size ", len(self._sendbuffer))

    def update(self):

        if not self._socket:
            raise ValueError("Not initialized")

        with self._lock:
            self._sendbuffer[0] = Sender.CMD_PAINT_PANEL
            for i, value in enumerate(self.panel.data):
                if self.panel.gamma_correction:
                    self._sendbuffer[i + 1] = Color.gammaCorrection(value)
                else:
                    """
                    Make shure, the CMD_PAINT_PANEL is not part of the color data
                    """
                    self._sendbuffer[i + 1] = (
                        value if value != Sender.CMD_PAINT_PANEL else value + 1
                    )

            try:
                self._socket.send(self._sendbuffer)
            except Exception as e:
                print("Could not send data via tcp. ", e)



from .sender import Sender
from .color import Color

import os


class ProgMemSender(Sender):
    def __init__(self, path=".", append=False):
        super().__init__()
        self._path = path
        self._append = append

    @property
    def filename(self):
        if self._append:
            return "{}.h".format(self.panel.id)

        return "{}_{:d}.h".format(self.panel.id, self.panel.frame)

    @property
    def path(self):
        """The path to store the progmem declarations. (readonly)
        """
        return self._path

    @property
    def size(self):
        """The number of bytes to be written for one definition, (readonly)
        """
        return len(self.panel._data)

    @property
    def name(self):
        """The name of the sprite. If append is TRue, then the name is a combination
        of the panel id and the frame number. Otherwise the name equals to the panel id.
        (readonly)
        """
        return "{}_{:d}".format(self.panel.id, self.panel.frame)

    def update(self):
        """Writes the paneldata to a c header file as a promemdefinition.
        If append is TRue, every frame is appended as a new sprite definition to 
        the created file. Otherwise for every frame a new file is created. The filename
        is a combination of panel id and framenumber.

        If gammacorrection is activated for the display, all bytes written to the file
        are gamma corrected. See :class:`ledwall.components.Display` for deatils about
        gamma correction.
        """
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        mode = (
            (os.O_WRONLY | os.O_APPEND | os.O_CREAT) if self._append else (os.O_WRONLY | os.O_CREAT)
        )
        fd = os.open(os.path.join(self.path, self.filename), mode)

        try:
            width = self.panel.columns
            height = self.panel.rows
            os.write(
                fd,
                "const uint8_t {}[] PROGMEM = {} 0x{:02x}, 0x{:02x},\n".format(
                    self.name, "{", width, height
                ),
            )
            for i in range(0, self.size, 8):
                if self.panel.gamma_correction:
                    rowdata = Color.gammaCorrection(self.panel._data[i : i + 8])
                else:
                    rowdata = self.panel._data[i : i + 8]

                bytes = ["0x{:02x}, ".format(b) for b in rowdata]
                os.write(fd, "    {}\n".format("".join(bytes)))
            os.write(fd, "0x00};\n\n")

        finally:
            os.close(fd)



from threading import Thread, Lock

from .sender import Sender


class AsyncSender(Sender):
    def __init__(self, delegate):
        """Asnchronous sender.
        This sende. takes another sender as a delegate and calls the corresponding update 
        method in a separate thread. Although it is possible, it is not advisable to 
        wrap an async sender in an async sender,
        """
        super().__init__()
        if not isinstance(delegate, Sender):
            raise ValueError("The delegate must implement the Sender interface.")
        self._delegate = delegate
        self._lock = Lock()

    def update(self):
        t = Thread(target=self.__delegate_update)
        t.start()

    def __delegate_update(self):
        if self._lock.acquire(False):
            try:
                self._delegate.update()
            finally:
                self._lock.release()
        else:
            print("dropped frame ", self.panel.frame)

    def init(self, panel):
        super().init(panel)
        self._delegate.init(panel)



from inputs import get_key
from .eventqueue import EventEmitter, Event


class KeyboardEmitter(EventEmitter):
    def __init__(self):
        super().__init__()

    def emit(self):
        try:
            events = get_key()
            for event in events:
                self.queue.put(
                    Event(
                        Event.KEYBOARD,
                        event.code,
                        {"ev_type": event.ev_type, "state": event.state, "code": event.code},
                    )
                )

        except Exception:
            pass



from ledwall.util import TimeDelta
from .eventqueue import EventEmitter, Event

import time
import serial


class SerialEmitter(EventEmitter):
    def __init__(self, port_name="/dev/ttyACM0", baudrate=115200):
        super().__init__()
        self.__buffer_size = 32
        self.__buffer = bytearray()
        self.__buffer_write_index = 0
        self._baudrate = baudrate
        self._port = port_name

        self._s = serial.Serial(self.port, self.baudrate)

    @property
    def baudrate(self):
        return self._baudrate

    @property
    def port(self):
        return self._port

    def emit(self):
        b = self._s.read()
        self.__buffer.extend(b)
        if ord(b) == 10:
            self.queue.put(Event(Event.SYSTEM, "log", {"message": self.__buffer.decode().strip()}))
            self.__buffer = bytearray()



from ..components.event import FramerateEmitter, EventDispatcher, GamepadEmitter, Event
from ..util import TimeDelta
from ..components import Display, WireMode


class Animation:
    def __init__(self, do_after=None):
        self._do_after = do_after

    def animate(self):
        pass

    def paint(self, display):
        pass


class Animations:
    def __init__(self):
        self._entries = []

    def __iadd__(self, a):
        self._entries.append(a)
        return self

    def __isub__(self, a):
        self._entries.remove(a)
        return self

    def __iter__(self):
        for a in self._entries:
            yield a

    def __len__(self):
        return len(self._entries)

    def animate(self):
        if len(self) == 0:
            return
        anims = self._entries[:]
        dead_anims = []
        for a in anims:
            if not a.animate():
                dead_anims.append(a)

        self._entries = [a for a in anims if a not in dead_anims]
        for a in dead_anims:
            if a._do_after:
                a._do_after()

    def paint(self, display):
        for a in self:
            a.paint(display)


class Application(object):
    def __init__(self, display, framerate):
        # type : (Display) -> None
        self._display = display
        self._display.framerate = None
        self._display.fill((0, 0, 0))
        self._display.update()
        self._framerate = framerate
        self._running = True
        self._event_dispatcher = EventDispatcher()
        if framerate and framerate > 0:
            self._event_dispatcher.add_emitter(FramerateEmitter(framerate))

        self._event_dispatcher.add_emitter(GamepadEmitter())
        self._animations = Animations()
        self._paint_function = self.paint

    @property
    def width(self):
        return self._display.columns

    @property
    def height(self):
        return self._display.rows

    @property
    def paint_function(self):
        return self._paint_function

    @paint_function.setter
    def paint_function(self, f):
        if not f:
            self._paint_function = self.paint
            return

        if not callable(f):
            raise ValueError("Provided value is not callable.")

        self._paint_function = f

    @property
    def animations(self):
        return self._animations

    @property
    def display(self):
        return self._display

    @property
    def frame(self):
        return self.display.frame

    def add_animation(self, animation):
        animation._app = self
        self._animations += animation

    def add_emitter(self, emitter):
        # self._event_dispatcher.add_emitter(emitter)
        self._event_dispatcher += emitter

    def paint(self):
        pass

    def handle_event(self, event):
        pass

    def start_loop(self):
        try:
            while self._running:
                event = self._event_dispatcher.next_event()
                if (event.type, event.action) == (Event.SYSTEM, "update"):
                    self.update()
                else:
                    self.handle_event(event)
        except KeyboardInterrupt:
            self.stop_loop()
            print("\nGood Buy")
            print("May the force be with you ...")

    def stop_loop(self):
        self._running = False
        self._event_dispatcher.stop()

    def update(self):
        self._paint_function()
        self._animations.animate()
        self._animations.paint(self._display)
        try:
            # self._event_dispatcher.suspend()
            self._display.update()
        finally:
            # self._event_dispatcher.resume()
            pass


class Game(Application):
    def __init__(self, sender, framerate, width=10, height=10, mode=WireMode.ZIGZAG):
        super().__init__(Display(width, height, sender, mode=mode), framerate)
        self.action_map = {
            (Event.GAMEPAD, "ABS_X", 0): self.btn_left_pressed,
            (Event.GAMEPAD, "ABS_X", 255): self.btn_right_pressed,
            (Event.GAMEPAD, "ABS_Y", 0): self.btn_top_pressed,
            (Event.GAMEPAD, "ABS_Y", 255): self.btn_down_pressed,
            (Event.GAMEPAD, "BTN_THUMB", 1): self.btn_a_pressed,
            (Event.GAMEPAD, "BTN_THUMB2", 1): self.btn_b_pressed,
            (Event.GAMEPAD, "BTN_PINKIE", 1): self.btn_r_pressed,
            (Event.GAMEPAD, "BTN_TOP2", 1): self.btn_l_pressed,
            (Event.GAMEPAD, "BTN_TRIGGER", 1): self.btn_x_pressed,
            (Event.GAMEPAD, "BTN_TOP", 1): self.btn_y_pressed,
            (Event.GAMEPAD, "BTN_BASE3", 1): self.btn_select_pressed,
            (Event.GAMEPAD, "BTN_BASE4", 1): self.btn_start_pressed,
            (Event.GAMEPAD, "ABS_Y", 127): self.btn_abs_y_released,
        }

    def handle_event(self, event):
        action = self.action_map.get((event.type, event.action, event["state"]), None)
        if action:
            action()

    def register_action(self, source, action_name, state, action):
        self.set_action(source, action_name, state, action)

    def unregister_action(self, source, action_name, state):
        try:
            self.action_map.pop((source, action_name, state))
        except KeyError:
            return None

    def set_action(self, source, action_name, state, action):
        old_action = self.action_map.get((source, action_name, state))
        self.action_map[(source, action_name, state)] = action
        return old_action

    def btn_a_pressed(self):
        pass

    def btn_b_pressed(self):
        pass

    def btn_x_pressed(self):
        pass

    def btn_y_pressed(self):
        pass

    def btn_r_pressed(self):
        pass

    def btn_l_pressed(self):
        pass

    def btn_up_pressed(self):
        pass

    def btn_down_pressed(self):
        pass

    def btn_top_pressed(self):
        pass

    def btn_abs_y_released(self):
        pass

    def btn_left_pressed(self):
        pass

    def btn_right_pressed(self):
        pass

    def btn_select_pressed(self):
        pass

    def btn_start_pressed(self):
        pass



from .display import WireMode, Display
from .sender import Sender


class RegionSender(Sender):
    """
        Takes a rectangular region of a display and uses it
        as an input to a chained sender instance.

        This allows it to define a bigger display than the actual
        physical device and map a region to the physical device.

        The delegate parameter is used as the sender to hand over
        the color data for the defined region. The dimensions of the 
        region and the size of the physical device, the sender is 
        transmitting the data to, must be identical. 

        In combinatoin with the :class:`~ledwall.components.ListSender`,
        it is possible to combine multiple physical devices to create a
        connected display.

        This sender is a middleware component. It extends the 
        :class:`~ledwall.components.Sender` class and therefore can 
        be used as a sender attribute in the 
        :class:`~ledwall.components.Display` class.

        For the delegate, this class mimics a display with the desired
        region.

        ..warning::
            This class exposes not every display method to the delegate.
            Just the properties and methods needed to send the data to
            the physical display.  

        :param int x: The x value of the top left corner of the region
        :param int y: The y value of the top left corner of the region
        :param int width: The width of the region
        :param int height: The height of the region
        
        :param delegate: The sender instance to send the color data to
        :type delegate: Sender

        :param mode: The mode that the LEDs are organized. 
            Left-to-Right or Zig-Zag. Defaults to WireMode.LTR.
        :type mode: WireMode
    """

    def __init__(self, x, y, width, height, delegate, mode=WireMode.LTR):
        self.delegate = delegate
        self.width = width
        self.height = height
        self.x = x
        self.y = y
        self._pixbufiter = None
        self.display = None
        self._mode = mode

    @property
    def data(self):
        if not self._pixbufiter:
            self._pixbufiter = PixBufIter(self, self.display)

        return self._pixbufiter

    @property
    def gamma_correction(self):
        return self.display._gamma_correction

    @property
    def count(self):
        return self.width * self.height

    @property
    def mode(self):
        return self._mode

    def init(self, panel):
        self.display = panel
        self.delegate.init(self)

    def update(self):
        self.delegate.update()


class PixBufIter:
    def __init__(self, parent, display):
        self.display = display
        self.parent = parent

    @property
    def mode(self):
        return self.parent.mode

    @property
    def width(self):
        return self.parent.width

    @property
    def height(self):
        return self.parent.height

    def __getitem__(self, key):
        x, y = key

        if self.mode == WireMode.ZIGZAG and Display.odd_row(y + self.parent.y):
            return self.display[(self.parent.x + (self.width - x - 1), self.parent.y + y)]

        return self.display[(self.parent.x + x, self.parent.y + y)]

    def __iter__(self):
        for dy in range(self.height):
            for dx in range(self.width):
                color = self[(dx, dy)]
                yield color[0]
                yield color[1]
                yield color[2]

    @property
    def count(self):
        return self.width * self.height

    def __len__(self):
        return self.count * 3



from ..geometry.rectangle import Rectangle

import random


def int_to_bitmask(i):
    result = 0
    for _ in range(i):
        result >>= 1
        result |= 0b10000000
    return result


def convert_byte(image, x, y, high_color=(0, 0, 0), mask=0xFF):
    if y >= image.height:
        return 0
    if x >= image.width:
        return 0

    result = 0
    for b in range(x, x + 8):
        result = result << 1
        if b < image.width:
            color = image.getpixel((b, y))
            if color == high_color:
                result |= 1

    return result & mask


def convert_image_to_progmem_bitmask(
    name, image, x, y, width=None, height=None, high_color=(0, 0, 0)
):
    if width is None:
        width = image.width
    if height is None:
        height = image.height
    number_of_blank_bits = width % 8
    print(number_of_blank_bits)
    result = "const uint8_t {}[] PROGMEM = {} 0x{:02x}, 0x{:02x},\n    ".format(
        name, "{", width, height
    )
    bit_pos = range(x, x + width, 8)

    for row in range(height):
        bytes = [convert_byte(image, x, row, high_color) for x in bit_pos]

        if number_of_blank_bits > 0:
            last_byte = bytes[-1:]
            last_byte[0] &= int_to_bitmask(number_of_blank_bits)
            bytes[-1:] = last_byte

        bytes = ["B{:08b}".format(b) for b in bytes]
        result += ", ".join(bytes)
        result += "\n    "

    result += "};"
    return result



"""
Copyright 2019

klaas.nebuhr@digicubes.de
"""
from typing import Tuple

from .point import Point
from ..util import intersectRect

TRect = Tuple(int, int, int, int)


class Rectangle(Point):
    @staticmethod
    def fromTuple(t: TRect):
        """
		Creates a new rectangle based on the provided four-value-tuple.
		The values of the tuple are:
		x, y, width, height.
		"""
        if len(t) < 4:
            raise ValueError("Cannot create rectangle from given value", t)
        return Rectangle(t[0], t[1], t[2], t[3])

    def __init__(self, x: int, y: int, width: int, height: int):
        super().__init__(x, y)
        self._width = width
        self._height = height

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, w: int):
        self._width = w

    @property
    def height(self) -> int:
        return self._height

    @height.setter
    def height(self, h: int):
        self._height = h

    @property
    def right(self) -> int:
        return self.x + self.width - 1

    @property
    def bottom(self) -> int:
        return self.y + self.height - 1

    @property
    def p1(self) -> Point:
        return Point(self.x, self.y)

    @property
    def p2(self) -> Point:
        return Point(self.right, self.y)

    @property
    def p3(self) -> Point:
        return Point(self.right, self.bottom)

    @property
    def p4(self) -> Point:
        return Point(self.x, self.bottom)

    @property
    def points(self) -> Tuple(Point, Point, Point, Point):
        return (self.p1, self.p2, self.p3, self.p4)

    def __str__(self):
        return "({:d},{:d},{:d},{:d})".format(self.x, self.y, self.width, self.height)

    def __repr__(self):
        return "Rectangle(x={:d},y={:d},width={:d},height={:d})".format(
            self.x, self.y, self.width, self.height
        )

    def __len__(self):
        return 4

    def __getitem__(self, key: int):
        return {
            0: self.x,
            1: self.y,
            2: self.width,
            3: self.height,
            4: self.right,
            5: self.bottom,
        }.get(key, None)

    def __iter__(self):
        yield self.x
        yield self.y
        yield self.width
        yield self.height

    def __sub__(self, other):
        return intersectRect(tuple(self), tuple(other))

    def __rsub__(self, other):
        return intersectRect(tuple(other), tuple(self))

    def __isub__(self, other):
        self.x, self.y, self.width, self.height = intersectRect(tuple(self), tuple(other))



from .timedelta import TimeDelta
from .settings import Settings


def intersectRect(r1, r2):

    x1 = max(r1[0], r2[0])
    x2 = min(r1[0] + r1[2], r2[0] + r2[2])
    dx = x2 - x1

    if dx < 0:
        return None

    y1 = max(r1[1], r2[1])
    y2 = min(r1[1] + r1[3], r2[1] + r2[3])
    dy = y2 - y1

    if dy < 0:
        return None
    return (x1, y1, dx, dy)


__all__ = ["intersectRect", "TimeDelta", "Settings"]



__all__ = ["Sender"]


class Sender(object):
    """
    The Sender is responsible for transferring the data
    from the display to the physical device. If a display has
    no sender, the :meth:`ledwall.components.Display.update`
    method will have no effect. In this case the display serves
    as a plane framebuffer. You can use is to layer several
    images and manipulate them independently.

    Currently the library supports a serial connection
    via the :class:`~ledwall.components.SerialSender`
    or a UDP based communication via the
    :class:`~ledwall.components.UDPSender`.
    """

    CMD_INIT_PANEL = 1
    CMD_CLEAR_PANEL = 2
    CMD_FILL_PANEL = 3
    CMD_PAINT_PANEL = 243
    CMD_SET_PIXEL = 5
    CMD_WRITE_RAW = 6
    CMD_SHOW = 255

    def __init__(self):
        self._panel = None

    @property
    def panel(self):
        """
        The associated display instance. Available
        after init has been called.
        """
        return self._panel

    def init(self, display):
        """
        Sets the associated display. This method will be called by
        the display, so don't call this method directly. When update() 
        is called, the implementation can access the display via the 
        property 'panel'. (I know, it's confusing and will be changed 
        in the future).
        """
        self._panel = display

    def update(self):
        """
        This method will be called every time, the sender should update
        the physical led display and should be implemented in every
        derived class.
        """
        pass



import threading
import queue

__ALL__ = "EventEmitter"


class Event(object):
    GAMEPAD = 1
    KEYBOARD = 2
    MOUSE = 3
    SYSTEM = 4
    MQTT = 5
    ARDUINO = 6
    USER = 999

    PRIORITY_HIGH = 1
    PRIORITY_NORMAL = 10
    PRIORITY_LOW = 100

    NAMES = {
        GAMEPAD: "GAMEPAD",
        KEYBOARD: "KEYBOARD",
        MOUSE: "MOUSE",
        SYSTEM: "SYSTEM",
        USER: "USER",
    }

    def __init__(self, event_type, action, data=None, priority=PRIORITY_NORMAL):
        """

        :param str action: The human readable event action
        :param dict data: Dictionary with additional type specific information
        :param int event_type: The 'source' of the event.
        """
        self._event_type = event_type
        self._data = data if data else {}
        self._action = action
        self._priority = priority

    @property
    def type(self):
        return self._event_type

    @property
    def data(self):
        return self._data

    @property
    def action(self):
        return self._action

    def state(self):
        return self.data.get("state", None)

    def __getitem__(self, key):
        return self._data.get(key, None)

    def __setitem__(self, key, value):
        self._data[key] = value

    def __str__(self):
        return self._action

    def __repr__(self):
        return "Event('{}','{}',{}, priority={})".format(
            Event.NAMES[self._event_type], self._action, self._data, self._priority
        )

    def __cmp__(self, other):
        return (self._priority > other._priority) - (self._priority < other._priority)

    def __lt__(self, other):
        return self._priority < other._priority


class EventDispatcher(object):
    def __init__(self):
        self._queue = queue.PriorityQueue()
        self._generators = []

    def add_emitter(self, src):
        self._generators.append(src)
        src.connect_queue(self._queue)

    def suspend(self):
        for g in self._generators:
            g.suspend()

    def resume(self):
        for g in self._generators:
            g.resume()

    def __iadd__(self, other):
        self.add_emitter(other)
        return self

    def next_event(self):
        """
        Returns the next event object. If no events are in the queue, the method blocks
        until a new event was added to the queue.

        :return: The event object.
        """
        return self._queue.get()

    def stop(self):
        for g in self._generators:
            g.stop()


class EventEmitter(threading.Thread):
    def __init__(self):
        super().__init__()
        self.queue = None
        self._is_running = False
        self.daemon = True
        self._pausing = False

    def connect_queue(self, q):
        self.queue = q
        self._is_running = True
        self.start()

    @property
    def suspended(self):
        return self._pausing

    def suspend(self):
        self._pausing = True

    def resume(self):
        self._pausing = False

    def stop(self):
        self._is_running = False

    def run(self):
        while self._is_running:
            if not self._pausing:
                self.emit()

    def emit(self):
        pass



import colorsys

# from .hsvcolor import HSVColor


class RGBColor(object):
    @staticmethod
    def fromIntValues(r, g, b):
        return RGBColor(r / 255.0, g / 255.0, b / 255.0)

    def __init__(self, r, g, b):
        self._r = float(r)
        if self._r < 0.0 or self._r > 1.0:
            raise ValueError("Red value out of range.", r)

        self._g = float(g)
        if self._g < 0.0 or self._g > 1.0:
            raise ValueError("Green value out of range.", g)

        self._b = float(b)
        if self._b < 0.0 or self._b > 1.0:
            raise ValueError("Green value out of range.", b)

    def __str__(self):
        return "({:.2f},{:.2f},{:.2f})".format(self.red, self.green, self.blue)

    def __repr__(self):
        return "RGBColor({:.2f},{:.2f},{:.2f})".format(self.red, self.green, self.blue)

    def __iter__(self):
        yield self._r
        yield self._g
        yield self._b

    @property
    def intValues(self):
        return (int(self.red * 255.0), int(self.green * 255.0), int(self.blue * 255.0))

    @property
    def red(self):
        return self._r

    @red.setter
    def red(self, value):
        self._r = value % 1.0

    @property
    def green(self):
        return self._g

    @green.setter
    def green(self, value):
        self._g = value % 1.0

    @property
    def blue(self):
        return self._b

    @blue.setter
    def blue(self, value):
        self._b = value % 1.0

    @property
    def hsv(self):
        return colorsys.rgb_to_hsv(self.red, self.green, self.blue)

    def mixin(self, color, factor=0.5):
        if factor < 0 or factor > 1:
            raise ValueError("Factor out of range.", factor)
        if isinstance(color, RGBColor):
            mefac = 1.0 - factor
            self._r = self._r * mefac + color._r * factor
            self._g = self._g * mefac + color._g * factor
            self._b = self._b * mefac + color._b * factor
        else:
            raise ValueError(
                "Unsupported data type for color. Expected instance of RGBColor", color
            )



import colorsys

from typing import Union, Tuple

from .hsvcolor import HSVColor
from .rgbcolor import RGBColor

__all__ = ["Color"]

gamma8_table = [
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    3,
    3,
    3,
    3,
    3,
    3,
    3,
    4,
    4,
    4,
    4,
    4,
    5,
    5,
    5,
    5,
    6,
    6,
    6,
    6,
    7,
    7,
    7,
    7,
    8,
    8,
    8,
    9,
    9,
    9,
    10,
    10,
    10,
    11,
    11,
    11,
    12,
    12,
    13,
    13,
    13,
    14,
    14,
    15,
    15,
    16,
    16,
    17,
    17,
    18,
    18,
    19,
    19,
    20,
    20,
    21,
    21,
    22,
    22,
    23,
    24,
    24,
    25,
    25,
    26,
    27,
    27,
    28,
    29,
    29,
    30,
    31,
    32,
    32,
    33,
    34,
    35,
    35,
    36,
    37,
    38,
    39,
    39,
    40,
    41,
    42,
    43,
    44,
    45,
    46,
    47,
    48,
    49,
    50,
    50,
    51,
    52,
    54,
    55,
    56,
    57,
    58,
    59,
    60,
    61,
    62,
    63,
    64,
    66,
    67,
    68,
    69,
    70,
    72,
    73,
    74,
    75,
    77,
    78,
    79,
    81,
    82,
    83,
    85,
    86,
    87,
    89,
    90,
    92,
    93,
    95,
    96,
    98,
    99,
    101,
    102,
    104,
    105,
    107,
    109,
    110,
    112,
    114,
    115,
    117,
    119,
    120,
    122,
    124,
    126,
    127,
    129,
    131,
    133,
    135,
    137,
    138,
    140,
    142,
    144,
    146,
    148,
    150,
    152,
    154,
    156,
    158,
    160,
    162,
    164,
    167,
    169,
    171,
    173,
    175,
    177,
    180,
    182,
    184,
    186,
    189,
    191,
    193,
    196,
    198,
    200,
    203,
    205,
    208,
    210,
    213,
    215,
    218,
    220,
    223,
    225,
    228,
    231,
    233,
    236,
    239,
    241,
    244,
    247,
    249,
    252,
    255,
]

TRGBColorTuple = Tuple[int, int, int]
TColor = Union[HSVColor, RGBColor, TRGBColorTuple]


class Color(object):
    """Represents a color on the display. Because of the nature of the WS2812b
    LED, colors are represented in the RGB color space. Each component is represented
    as a byte value [0;255].

    Color instances can be created directly or can be converted from :class:`RGBColor` instances
    or :class:`HSVColor` instances.

    This class provides a class method for gamacorrection. The gamma correction table
    is especially made for the WS2812b LEDs.
    """

    @staticmethod
    def convert(val):
        if isinstance(val, Color):
            return val
        if isinstance(val, str):
            return Color.fromHexString(val)
        if isinstance(val, RGBColor):
            return Color.fromRGBColor(val)
        if isinstance(val, HSVColor):
            return Color.fromHSVColor(val)
        if len(val) == 3:
            return Color.fromTuple(val)
        raise ValueError("Cannot convert value in color object")

    @staticmethod
    def gammaCorrection(val):
        """Maps the provided values to the corresponding gamma corrected
        values. See also the property :meth:`~ledwall.components.Color.gamma8`.

        .. code-block:: python

            from ledwall.components import *

            gamma8_table = [ Color.gamma_correction(c) for c in range(256) ]

            color  = Color(244,5,54)
            gcolor = [ Color.gamma_correction(c) for c in color ]

            garr = Color.gamma_correction([245,23,16,47,3,89,167,213])

        :param val: The value or the values to be corrected.
        :type val: (int or iterable(int))
        """
        if isinstance(val, int) or len(val) == 1:
            return gamma8_table[val]
        return [gamma8_table[v] for v in val]

    @staticmethod
    def fromHSVColor(color):
        """Class method to create a new color instance from an 
        existing HSVColor instance.

        :param color: The HSVColor instance
        :type color: HSVColor
        :rtype: Color
        """
        rgb = color.rgb
        return Color.fromRGBColor(RGBColor(rgb[0], rgb[1], rgb[2]))

    @staticmethod
    def fromRGBColor(color):
        return Color.fromTuple(color.intValues)

    @staticmethod
    def fromTuple(t):
        """Creates an instance from the first three values of the provided tuple.
        The tree values must be ints and within the range [0;255]. Color order is
        RGB.

        Example:

        .. code-block:: python
    
            from ledwall.components import *

            c = Color.fromTuple((255,12,234))

        :param tuple(int) t: A zubple with the RGB channel values.
        :rtype: Color
        """
        return Color(t[0], t[1], t[2])

    @staticmethod
    def fromHexString(color):
        """Creates a color instance from an css color string in hexadicimal
        notation. 

        Example:
        
        .. code-block:: python
    
            from ledwall.components import *

            c = Color.fromHexString('#3F234A')

        :param str color: The hex string notation of the collor.
        :rtype: Color
        """
        s = color.lstrip("#")
        return Color(int(s[0:2], 16), int(s[2:4], 16), int(s[4:6], 16))

    def __init__(self, r=0, g=0, b=0):
        """Creates a new color instance. If you don't provide a value
        for the r, g or b component, it defaults to zero. All values
        have to be int values and within the range of [0;255]
        """
        self.red = r
        self.green = g
        self.blue = b

    def __iter__(self):
        yield self._r
        yield self._g
        yield self._b

    def __repr__(self):
        return "Color(%d,%d,%d)" % (self.red, self.green, self.blue)

    def __str__(self):
        return self.hexStr

    def __eq__(self, other):
        if isinstance(other, Color):
            return other.red == self.red and other.green == self.green and other.blue == self.blue
        if (isinstance(other, tuple) or isinstance(other, list)) and len(other) == 3:
            return self.red == other[0] and self.green == other[1] and self.blue == other[2]
        return NotImplemented

    @property
    def red(self):
        """Returns the red component of this color.

        :rtype: int
        """
        return self._r

    @red.setter
    def red(self, value):
        """Sets the red value of this color.

        The provided value must be within the range [0;255]
        
        :param int value: The red channel value.
        """
        self._r = int(value)
        if self._r < 0 or self._r > 255:
            raise ValueError(
                "Only values between 0 and 255 are accepted for the red channel", value
            )

    @property
    def green(self):
        """Returns the green component of this color.
        
        :rtype: int
        """
        return self._g

    @green.setter
    def green(self, value):
        """Sets the green value of this color.

        The provided value must be within the range [0;255]
        
        :param int value: The green channel value.
        """
        self._g = int(value)
        if self._g < 0 or self._g > 255:
            raise ValueError(
                "Only values between 0 and 255 are accepted for the green channel", value
            )

    @property
    def blue(self):
        """Returns the blue component of this color.

        :rtype: int
        """
        return self._b

    @blue.setter
    def blue(self, value):
        """Sets the blue value of this color.

        The provided value must be within the range [0;255]
        
        :param int value: The blue channel value.
        """
        self._b = int(value)
        if self._b < 0 or self._b > 255:
            raise ValueError(
                "Only values between 0 and 255 are accepted for the blue channel", value
            )

    @property
    def hexStr(self):
        """Property for hexadecimal string representation of this color with a leading #. 
        This is an R/W property

        For Example:

        .. code-block:: python
    
            Color c = Color(255,0,16)
            print c.hexStr

            # This will print '#FF0010' to the console

            # Example for setting a color value
            c.hexStr = '#ffcc23'

        """
        return "#%0.2X%0.2X%0.2X" % (self._r, self._g, self._b)

    @hexStr.setter
    def hexStr(self, value):
        """
        Sets the RGB channel values via a hexstring. The hex string has
        to be well formatted. An exmple for a well formed hex string is
        '#fa345c'.

        :param str value: The hexstring
        """
        c = Color.fromHexString(value)
        self._r = int(c.red)
        self._g = int(c.green)
        self._b = int(c.blue)

    @property
    def floatValues(self):
        """Tuple of float values for the RGB channel.
        """
        return (self.red / 255.0, self.green / 255.0, self.blue / 255.0)

    @property
    def hsvColor(self):
        """Property for getting a converted HSVColor instance
        or updating the RGB channels by a HSVColor instance.
        """
        hsv = self.rgbColor.hsv
        return HSVColor(hsv[0], hsv[1], hsv[2])

    @hsvColor.setter
    def hsvColor(self, color):
        if not isinstance(color, HSVColor):
            raise ValueError("Needs a HSVColor instance as parameter", color)
        c = Color.fromHSVColor(color)
        self.red = c.red
        self.green = c.green
        self.blue = c.blue

    @property
    def rgbColor(self):
        """Converted RGBColor instance
        """
        return RGBColor.fromIntValues(self.red, self.green, self.blue)

    @property
    def gamma8(self):
        """Tuple of gamma8 corrected RGB values.
        """
        return (
            Color.gammaCorrection(self.red),
            Color.gammaCorrection(self.green),
            Color.gammaCorrection(self.blue),
        )

    def __getitem__(self, key):
        if isinstance(key, str):
            if key == "red" or key == "r":
                return self.red
            if key == "green" or key == "g":
                return self.green
            if key == "blue" or key == "b":
                return self.blue
            raise ValueError("Uknown string identifier to lookup item", key)

        arr = [c for c in self]
        return arr[key]



from .display import Display, WireMode
from .color import Color
from .colortable import ColorTable
from .hsvcolor import HSVColor
from .rgbcolor import RGBColor
from .serialsender import SerialSender
from .mqttsender import MqttSender
from .asyncsender import AsyncSender
from .consolesender import ConsoleSender
from .sender import Sender
from .progmemsender import ProgMemSender
from .listsender import ListSender
from .udpsender import UDPSender
from .tcpsender import TCPSender
from .application import Application, Game, Animation

from .multidisplay import RegionSender

__all__ = [
    "Color",
    "Display",
    "ColorTable",
    "RGBColor",
    "HSVColor",
    "SerialSender",
    "MqttSender",
    "Sender",
    "ConsoleSender",
    "AsyncSender",
    "ProgMemSender",
    "ListSender",
    "UDPSender",
    "TCPSender",
    "Application",
    "LEDWallError",
    "TransmissionError",
    "CoordinateError",
    "Game",
    "Animation",
    "WireMode",
    "RegionSender",
]

name = "ledwall"


class LEDWallError(Exception):
    pass


class TransmissionError(LEDWallError):
    pass


class CoordinateError(LEDWallError):
    pass


def rgb_to_hsv(val):
    return RGBColor.fromIntValues(val[0], val[1], val[2]).hsv



from .rectangle import Rectangle
from .point import Point


class Line(object):
    @staticmethod
    def fromTuple(t):
        if len(t) < 4:
            raise ValueError("Cannot create line from given value", t)
        return Line(Point(t[0], t[1]), Point(t[2], t[3]))

    def __init__(self, p1, p2):
        if not isinstance(p1, Point):
            raise ValueError("Expected point instance for p1")

        if not isinstance(p2, Point):
            raise ValueError("Expected point instance for p2")

        self._p1 = p1
        self._p2 = p2

    @property
    def p1(self):
        return self._p1

    @property
    def p2(self):
        return self._p2

    @property
    def bounds(self):
        minx = min(self._p1.x, self._p2.x)
        miny = min(self._p1.y, self._p2.y)
        maxx = max(self._p1.x, self._p2.x)
        maxy = max(self._p1.y, self._p2.y)

        return Rectangle(minx, miny, maxx - minx + 1, maxy - miny + 1)

    def __str__(self):
        return "({0!s},{1!s})".format(self.p1, self.p2)

    def __repr__(self):
        return "Line({0!r},{1!r})".format(self.p1, self.p2)

    def __iadd__(self, pdelta):
        self._p1 += pdelta
        self._p2 += pdelta

    def __isub__(self, pdelta):
        self._p1 -= pdelta
        self._p2 -= pdelta

    def paint(self, display, color=(255, 255, 255), update=False):
        # TODO: Not yet implemented
        raise NotImplementedError("Painting lines is not implpmented yet.")

    def __iter__(self):
        yield self.p1
        yield self.p2



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ledwall",
    version="1.0.11",
    author="Klaas Nebuhr (FirstKlaas)",
    author_email="klaas@nebuhr.de",
    description="A simple but powerful library to control WS2812b LED panel.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FirstKlaas/LEDWall/tree/master/python",
    packages=setuptools.find_packages(),
    install_requires=[
        'pyserial',
        'inputs',
        'paho-mqtt'
    ],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import colorsys

# from .rgbcolor import RGBColor


class HSVColor(object):
    """Represents a color in the HSV color space.
    The components of the HSV color are normalized to the
    intervall [0.0;1.0].

    If you have in values to create a HSVColor you can use
    the class method fromIntValues to create an instance.

    The hue component has to be within the range of [0;360]. Saturation
    and value in the range of [0;100]

    The rgb property returns the converted and normalized values in the
    RGB color space.

    """

    @staticmethod
    def fromIntValues(h, s, v):
        return HSVColor(h / 360.0, s / 100.0, v / 100.0)

    def __init__(self, h=0.0, s=1.0, v=1.0):
        """Creates a new color instance in the HSV color space. 
        The values for hue, value and saturation have to be provided 
        in normalized [0.0;1.0] form.
        """
        self._h = h
        self._v = v
        self._s = s

    @property
    def hue(self):
        """The normalized hue component of the color.
        
        :rtype: float
        """
        return self._h

    @hue.setter
    def hue(self, value):
        self._h = value % 1.0

    @property
    def hue360(self):
        return self.hue * 360.0

    @hue360.setter
    def hue360(self, value):
        self.hue = (value / 360.0) % 1.0

    @property
    def h(self):
        """Same as the property hue. Just for the lazy people.
        
        :rtype: float
        """
        return self.hue

    @h.setter
    def h(self, value):
        self.hue = value

    @property
    def saturation(self):
        """The normalized saturation component of the color.
        
        :rtype: float
        """
        return self._s

    @saturation.setter
    def saturation(self, val):
        self.s = val % 1.0

    @property
    def saturation100(self):
        """The saturation component of the color as a value 
           in the intervall of [0.0;100.0].
        
        :rtype: float
        """
        return self._s * 100.0

    @saturation100.setter
    def saturation100(self, val):
        self.s = (val / 100.0) % 1.0

    @property
    def s(self):
        """Same as the property saturation. Just for the lazy people.
        
        :rtype: float
        """
        return self.saturation

    @s.setter
    def s(self, val):
        self.saturation = val

    @property
    def value(self):
        """The normalized value component of the color.
        
        :rtype: float
        """
        return self._v

    @value.setter
    def value(self, val):
        self._v = (val % 1.0) if val > 1.0 else val

    @property
    def value100(self):
        """The value component of the color in the range
           of [0.0;100.0].
        
        :rtype: float
        """
        return self.value * 100

    @value100.setter
    def value100(self, val):
        self.value = (val / 100.0) % 1.0

    @property
    def v(self):
        """Same as the property value. Just for the lazy people.
        
        :rtype: float
        """
        return self.value

    @v.setter
    def v(self, val):
        self.value = val

    @property
    def intValues(self):
        """Returns the int values for the three components. The
        normalized hue value is projected on the intervall [0;360]
        and value and saturation are projeted on the intervall
        [0;100]

        :rtype: tuple(int)
        """
        return (
            int(round(self.hue * 360.0)),
            int(round(self.saturation * 100.0)),
            int(round(self.value * 100.0)),
        )

    def __iter__(self):
        yield self.h
        yield self.s
        yield self.v

    def __repr__(self):
        return "HSVColor({:.2f},{:.2f},{:.2f})".format(self.h, self.s, self.v)

    def __str__(self):
        return "({:.2f},{:.2f},{:.2f})".format(self.h, self.s, self.v)

    def __getitem__(self, key):
        if isinstance(key, str):
            if key == "hue" or key == "h":
                return self.h
            if key == "saturation" or key == "s":
                return self.s
            if key == "value" or key == "v":
                return self.v
            raise ValueError("Uknown string identifier to lookup item", key)

        elif isinstance(key, int):
            if key < 0 or key > 2:
                raise ValueError("Index out ouf bounds [0,2]", key)

        elif isinstance(key, slice):
            if abs(key.start) > 2:
                raise ValueError("Slice start ouf bounds [-2,2]", key)

        raise ValueError("Unsupported index type")

    @property
    def rgb(self):
        """Returns a tuple with the converted and normalized rgb values

        :rtype: tuple(float)
        """
        return colorsys.hsv_to_rgb(self.hue, self.saturation, self.value)



"""
The classic tetris game ported to the LED Wall
"""
import random

import ledwall.components as comp

TEMPLATEWIDTH = 5
TEMPLATEHEIGHT = 5

S_SHAPE_TEMPLATE = [
    [".....", ".....", "..OO.", ".OO..", "....."],
    [".....", "..O..", "..OO.", "...O.", "....."],
]

Z_SHAPE_TEMPLATE = [
    [".....", ".....", ".OO..", "..OO.", "....."],
    [".....", "..O..", ".OO..", ".O...", "....."],
]

I_SHAPE_TEMPLATE = [
    ["..O..", "..O..", "..O..", "..O..", "....."],
    [".....", ".....", "OOOO.", ".....", "....."],
]

O_SHAPE_TEMPLATE = [[".....", ".....", ".OO..", ".OO..", "....."]]

P_SHAPE_TEMPLATE = [[".....", ".....", "..O..", ".....", "....."]]

J_SHAPE_TEMPLATE = [
    [".....", ".O...", ".OOO.", ".....", "....."],
    [".....", "..OO.", "..O..", "..O..", "....."],
    [".....", ".....", ".OOO.", "...O.", "....."],
    [".....", "..O..", "..O..", ".OO..", "....."],
]

L_SHAPE_TEMPLATE = [
    [".....", "...O.", ".OOO.", ".....", "....."],
    [".....", "..O..", "..O..", "..OO.", "....."],
    [".....", ".....", ".OOO.", ".O...", "....."],
    [".....", ".OO..", "..O..", "..O..", "....."],
]

T_SHAPE_TEMPLATE = [
    [".....", "..O..", ".OOO.", ".....", "....."],
    [".....", "..O..", "..OO.", "..O..", "....."],
    [".....", ".....", ".OOO.", "..O..", "....."],
    [".....", "..O..", ".OO..", "..O..", "....."],
]

C_SHAPE_TEMPLATE = [
    [".....", "..O..", "..OO.", ".....", "....."],
    [".....", ".....", "..OO.", "..O..", "....."],
    [".....", ".....", ".OO..", "..O..", "....."],
    [".....", "..O..", ".OO..", ".....", "....."],
]

D_SHAPE_TEMPLATE = [
    [".....", "..O..", "..O..", ".....", "....."],
    [".....", ".....", ".OO..", ".....", "....."],
]

BLANK = "."

PIECES = {
    "S": S_SHAPE_TEMPLATE,
    "Z": Z_SHAPE_TEMPLATE,
    "I": I_SHAPE_TEMPLATE,
    "J": J_SHAPE_TEMPLATE,
    "L": L_SHAPE_TEMPLATE,
    "O": O_SHAPE_TEMPLATE,
    "T": T_SHAPE_TEMPLATE,
    "P": P_SHAPE_TEMPLATE,
    "C": C_SHAPE_TEMPLATE,
    "D": D_SHAPE_TEMPLATE,
}

PIECES_COLOR = {
    "S": (47, 20, 201),
    "Z": (47, 20, 201),
    "I": (47, 20, 201),
    "J": (47, 20, 201),
    "L": (47, 20, 201),
    "O": (47, 20, 201),
    "T": (47, 20, 201),
    "P": (47, 20, 201),
    "C": (47, 20, 201),
    "D": (47, 20, 201),
    BLANK: (0, 0, 0),
    "B": (255, 0, 0),
}


"""
PIECES_COLOR = {'S': (255, 0, 0),
                'Z': (255, 255, 0),
                'I': (0, 255, 0),
                'J': (0, 255, 255),
                'L': (255, 0, 255),
                'O': (0, 0, 255),
                'T': (255, 100, 255),
                'P': (86, 244, 66),
                'C': (255, 249, 226),
                'D': (152, 47, 175),
                BLANK: (0, 0, 0),
                'B': (255, 255, 255),
                }
"""

# PIECES_ORDER = {'S': 0,'Z': 1,'I': 2,'J': 3,'L': 4,'O': 5,'T': 6}


class Tetris(object):
    """
    The Tetris Game Class.
    """

    OVERFLOW_LEFT = 0
    OVERFLOW_RIGHT = 1
    OVERFLOW_BOTTOM = 2
    OVERFLOW_TOP = 3
    COLLISION = 4
    VALID_POSITION = 5

    def __init__(self, display):
        self._display = display
        self._currentPiece = self.get_new_piece()
        self._matrix = [[BLANK for x in range(display.columns)] for y in range(display.rows)]
        self._falling_speed = 0.1

    @property
    def piece(self):
        """The current piece that drops down."""
        return self._currentPiece

    @property
    def matrix(self):
        """The matrix"""
        return self._matrix

    @property
    def width(self):
        """Number of columns"""
        return self._display.columns

    @property
    def height(self):
        """The number of rows."""
        return self._display.rows

    def set_block_color(self, block_id, color):
        """
        Sets the background color for the specified
        block to the given color.
        """
        if block_id not in PIECES.keys():
            raise ValueError("Uknown block id")
        PIECES_COLOR[block_id] = tuple(comp.Color.convert(color))

    def set_background_color(self, color):
        """
        Sets the background color for the blank
        pixels.
        """
        PIECES_COLOR[BLANK] = tuple(comp.Color.convert(color))

    @property
    def block_ids(self):
        """The block ids"""
        return PIECES.keys()

    def fallDown(self):
        """
        No clue.
        """
        # TODO: Was ist das hier?
        if self._currentPiece is None:
            return False

    def clearMatrix(self):
        """
        Clears the matrix. Every pixel in the
        matrix is set to BLANK.
        """
        for x in range(self.width):
            for y in range(self.height):
                self.matrix[x][y] = BLANK

    def isRowComplete(self, row):
        """
        Test, wether the given row is blank
        or not. Returns `true`, if all
        the pixels in the given row are
        blank. `False` else.
        """
        for x in range(self.width):
            if self.matrix[x][row] == BLANK:
                return False
        return True

    def isColumnComplete(self, column):
        """
        Test, wether the given column is blank
        or not. Returns `true`, if all
        the pixels in the given column are
        blank. `False` else.
        """
        if column < 0 or column >= self.width:
            raise ValueError("Column value out of range")

        for y in range(self.height):
            if self.matrix[column][y] == BLANK:
                return False
        return True

    def dissolveColumn(self, column):
        for y in range(self.height):
            self._matrix[column][y] = "B"

        for val in range(255, 0, -25):
            # Set the color for the column
            PIECES_COLOR["B"] = (val, val, val)
            self.writeMatrixToDisplay()
            # self._display.update()

        PIECES_COLOR["B"] = [255, 255, 255]

    def getCompletedColumns(self):
        result = []
        for x in range(self.width):
            if self.isColumnComplete(x):
                result += [x]
        return result

    def getCompletedRows(self):
        result = []
        for y in range(self.height):
            if self.isRowComplete(y):
                result += [y]
        return result

    def deleteRow(self, row):
        # Move rows down.
        for y in range(row - 1, -1, -1):
            for x in range(self.width):
                self._matrix[x][y + 1] = self._matrix[x][y]

        # Insert a new blank row at he top.
        for x in range(self.width):
            self._matrix[x][0] = BLANK

    def deleteColumn(self, column):
        for y in range(self.height):
            self._matrix[column][y] = BLANK

    def deleteCompleteRows(self):
        deletedRows = 0
        for y in range(self.height - 1, -1, -1):
            while self.isRowComplete(y):
                deletedRows += 1
                self.deleteRow(y)
        return deletedRows

    def deleteCompleteColumns(self):
        for c in self.getCompletedColumns():
            self.deleteColumn(c)

    def getShape(self, piece=None):
        piece = self.__ensurePiece(piece)

        return PIECES[piece["shape"]][piece["rotation"]]

    def update(self):
        self.writeMatrixToDisplay()
        if self._currentPiece:
            self.drawPiece(self._currentPiece)
        # self._display.update()

    def writePieceToMatrix(self, piece=None):
        piece = self.__ensurePiece(piece)

        shape = self.getShape(piece)
        for x in range(TEMPLATEWIDTH):
            for y in range(TEMPLATEHEIGHT):
                if shape[y][x] != BLANK:
                    mx = x + piece["x"]
                    my = y + piece["y"]
                    if not (not (0 <= mx < self.width) or not (0 <= my < self.height)):
                        self.matrix[mx][my] = piece["shape"]

    def writeMatrixToDisplay(self):
        for x in range(self.width):
            for y in range(self.height):
                shape = self.matrix[x][y]
                self._display.set_pixel(x, y, PIECES_COLOR[shape])

    def setBrickAt(self, x, y):
        self.matrix[x][y] = "B"

    def rotateCW(self, piece=None):
        piece = self.__ensurePiece(piece)

        shape = piece["shape"]
        piece["rotation"] += 1
        piece["rotation"] %= len(PIECES[shape])

    def rotateCCW(self, piece=None):
        piece = self.__ensurePiece(piece)

        shape = piece["shape"]
        piece["rotation"] += len(PIECES[shape])
        piece["rotation"] -= 1
        piece["rotation"] %= len(PIECES[shape])

    def new_piece(self):
        self._currentPiece = self.get_new_piece()

    def get_new_piece(self):
        # return a random new piece in a random rotation and color
        shape = random.choice(list(PIECES.keys()))
        new_piece = {
            "shape": shape,
            "rotation": random.randint(0, len(PIECES[shape]) - 1),
            "x": int(self.width / 2) - int(TEMPLATEWIDTH / 2),
            "y": -3,  # start it above the board (i.e. less than 0)
            "color": (124, 218, 255),
        }
        return new_piece

    def checkForCollision(self, piece=None, dx=0, dy=0):
        piece = self.__ensurePiece(piece)

        shapeToTest = self.getShape(piece)
        for x in range(TEMPLATEWIDTH):
            for y in range(TEMPLATEHEIGHT):
                mx = x + piece["x"] + dx
                my = y + piece["y"] + dy
                if mx >= 0 and mx < self.width and my >= 0 and my < self.height:
                    if shapeToTest[y][x] != BLANK and self.matrix[mx][my] != BLANK:
                        return Tetris.COLLISION
        return Tetris.VALID_POSITION

    def isOnBoard(self, piece=None, dx=0, dy=0):
        """Test if the bounds of the piece are on the
           display. It is ok that the piece extends the upper
           border.
        """
        piece = self.__ensurePiece(piece)
        x = piece["x"] + dx
        y = piece["y"] + dy

        return x >= 0 and x + TEMPLATEWIDTH < self.width and y + TEMPLATEHEIGHT < self.height

    def __ensurePiece(self, piece):
        if piece is None:
            piece = self._currentPiece
        if piece is None:
            raise ValueError("Provided pice is None and we have no 'currentPiece'")
        return piece

    def moveX(self, piece=None, val=0):
        piece = self.__ensurePiece(piece)
        piece["x"] += val

    def testOverflowX(self, piece=None, dx=0, dy=0):
        piece = self.__ensurePiece(piece)
        shapeToTest = self.getShape(piece)

        px = piece["x"] + dx
        # py = piece['y'] + dy

        for x in range(TEMPLATEWIDTH):
            for y in range(TEMPLATEHEIGHT):
                if px + x < 0 and shapeToTest[y][x] != BLANK:
                    return Tetris.OVERFLOW_LEFT

                elif px + x >= self.width and shapeToTest[y][x] != BLANK:
                    return Tetris.OVERFLOW_RIGHT
        return Tetris.VALID_POSITION

    def testOverflowY(self, piece=None, dx=0, dy=0):
        piece = self.__ensurePiece(piece)
        shapeToTest = self.getShape(piece)

        # px = piece['x'] + dx
        py = piece["y"] + dy

        for x in range(TEMPLATEWIDTH):
            for y in range(TEMPLATEHEIGHT):
                if py + y >= self.height and shapeToTest[y][x] != BLANK:
                    return Tetris.OVERFLOW_BOTTOM
                if py + y < 0 and shapeToTest[y][x] != BLANK:
                    return Tetris.OVERFLOW_TOP

        return Tetris.VALID_POSITION

    def isValidPosition(self, piece=None, dx=0, dy=0):
        piece = self.__ensurePiece(piece)
        if self.isOnBoard(piece, dx, dy):
            return not self.checkForCollision(piece, dx, dy)

        shapeToTest = self.getShape(piece)
        px = piece["x"] + dx
        py = piece["y"] + dy

        for x in range(TEMPLATEWIDTH):
            for y in range(TEMPLATEHEIGHT):
                if px + x < 0 and shapeToTest[y][x] != BLANK:
                    return Tetris.OVERFLOW_LEFT

                elif px + x >= self.width and shapeToTest[y][x] != BLANK:
                    return Tetris.OVERFLOW_RIGHT

                if py + y >= self.height and shapeToTest[y][x] != BLANK:
                    return Tetris.OVERFLOW_BOTTOM

                if shapeToTest[y][x] != BLANK and self.matrix[px + x][py + y] != BLANK:
                    return Tetris.COLLISION
        return Tetris.VALID_POSITION

    def copyRow(self, src, dst):
        for x in range(self.width):
            self.matrix[x][dst] = self.matrix[x][src]

    def drawPiece(self, piece=None, pixelx=None, pixely=None):
        piece = self.__ensurePiece(piece)
        shapeToDraw = self.getShape(piece)
        if pixelx == None and pixely == None:
            # if pixelx & pixely hasn't been specified, use the location stored in the piece data structure
            pixelx = piece["x"]
            pixely = piece["y"]

        # draw each of the boxes that make up the piece
        for x in range(TEMPLATEWIDTH):
            for y in range(TEMPLATEHEIGHT):
                if shapeToDraw[y][x] != BLANK:
                    px = x + pixelx
                    py = y + pixely
                    if py >= 0 and py < self.height and px >= 0 and py < self.width:
                        self._display.set_pixel(px, py, piece["color"])



from __future__ import division
import datetime


class TimeDelta(object):
    def __init__(self):
        self._max = None
        self._min = None
        self._millis = None
        self._begin = None
        self._diff = None

    @property
    def hasStarted(self):
        return self._begin is not None

    def begin(self):
        self._begin = datetime.datetime.now()

    def measure(self):
        if not self.hasStarted:
            return None

        self._diff = datetime.datetime.now() - self._begin
        self._millis = self._diff.total_seconds() * 1000
        self._min = min(self._min, self._millis) if self._min else self._millis
        self._max = max(self._max, self._millis) if self._max else self._millis
        return self._millis, self._min, self._max

    @property
    def millis(self):
        return self._millis

    @property
    def micros(self):
        return self._diff.microseconds if self._diff else None

    @property
    def seconds(self):
        return self._diff.seconds if self._diff else None

    @property
    def days(self):
        return self._diff.days if self._diff else None

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    def _NoneOrStr(self, val):
        return "%.2f" % val if val else "-"

    def asTuple(self):
        return (self.millis, self.min, self.max)

    def __str__(self):
        return "(ms: %s, min: %s, max: %s)" % (
            self._NoneOrStr(self._millis),
            self._NoneOrStr(self._min),
            self._NoneOrStr(self._max),
        )



from .sender import Sender
from .asyncsender import AsyncSender


class ListSender(Sender):
    """Manages a list of senders, so multiple panelscan be updated 
    by just one display. If async is true, all senders in the delegates
    list are wrapped in an :class:`~ledwall.components.AsyncSender` instance.
    Otherwise the init and update methods are called for each element of
    the list.

    :param delegates: An iterable object of sender instances.
    :type delegates: iterable(Sender)
    :param boolean async: Calls sender asynchronously if True. Directly else.
    """

    def __init__(self, delegates=None, add_async=False):
        super().__init__()
        self._add_async = add_async
        if delegates:
            if add_async:
                self._delegates = [AsyncSender(s) for s in delegates]
            else:
                self._delegates = delegates
        else:
            self._delegates = []

    def __iadd__(self, other):
        if self._add_async:
            self._delegates.append(AsyncSender(other))
        else:
            self._delegates.append(other)
        return self

    def init(self, panel):
        """Calls init for every provided sender.
        """
        super().init(panel)
        for s in self._delegates:
            s.init(self.panel)

    def update(self):
        """Calls update for every provided sender.
        """
        for s in self._delegates:
            s.update()
